@@ -16,6 +16,7 @@
 	with this program; if not, write to the Free Software Foundation, Inc.,
 	51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */
+#include "config.h"
 #define _DEFAULT_SOURCE
 #include <libintl.h>
 #include <locale.h>
@@ -51,9 +52,16 @@
 #include <linux/if_ether.h>
 #include <sys/mman.h>
 #endif
-#include <config.h>
-
 #include <openssl/evp.h>
+
+#if (HAVE_READPASSPHRASE == 1)
+#include <readpassphrase.h>
+#elif (HAVE_BSDREADPASSPHRASE == 1)
+#include <bsd/readpassphrase.h>
+#else
+#warning "Falling back to getpass(3), which is marked obsolete!"
+#endif
+
 #include "mtwei.h"
 #include "protocol.h"
 #include "console.h"
@@ -623,7 +631,8 @@ int main(int argc, char **argv) {
 					  "                 root privileges.\n"
 					  "  -a <path>      Use specified path instead of the default: %s for autologin config file.\n"
 					  "  -A             Disable autologin feature.\n"
-					  "  -t <timeout>   Amount of seconds to wait for a response on each interface.\n"
+					  "  -t <timeout>   Amount of seconds to wait for a successufl connection, or amount of seconds\n"
+					  "                 to wait for MNDP response (when used with -l).\n"
 					  "  -u <user>      Specify username on command line.\n"
 					  "  -p <password>  Specify password on command line.\n"
 					  "  -U <user>      Drop privileges to this user. Used in conjunction with -n\n"
@@ -717,8 +726,12 @@ int main(int argc, char **argv) {
 	}
 
 	if (!have_password) {
-		char *tmp;
-		tmp = getpass(quiet_mode ? "" : _("Password: "));
+#if (HAVE_READPASSPHRASE == 1 || HAVE_BSDREADPASSPHRASE == 1)
+		static char pwd[200];
+		char *tmp = readpassphrase(_("Password: "), (char *)&pwd, 200, RPP_ECHO_OFF);
+#else
+		char *tmp = getpass(_("Password: "));
+#endif
 #if defined(_POSIX_MEMLOCK_RANGE) && _POSIX_MEMLOCK_RANGE > 0
 		mlock(password, sizeof(password));
 #endif
